/*--

THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.

Copyright (C) 1999  Microsoft Corporation.  All rights reserved.

Module Name:

    person.c

Abstract:

    Utility functions for creating, finding, and deleting a user account
    using LDAP.

--*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#include <netdb.h>
#include <sys/param.h>
#include <krb5.h>
#include "lber.h"
#include "ldap.h"
#include "sam.h"
#include "person.h"

static const char rcsid[] = "$Id$";

int find_user(
    LDAP *ld,
    LDAPMessage **res,
    char *bind_path,
    char *user
    )
{
    char search_exp[1024];

    sprintf(search_exp, "(samAccountName=%s)", user);
    return ldap_search_s(ld, bind_path, LDAP_SCOPE_SUBTREE,
			 search_exp, NULL, 0, res);
}

int add_user(
    LDAP *ld,
    char *bind_path,
    char *user,
    char *domainname,
    int invalidate,
    char **optattrNames,
    char **optattrValues
    )
{
#define ADD_ATTR(t, v) 		\
    mods[n] = malloc(sizeof(LDAPMod));	\
    mods[n]->mod_op = LDAP_MOD_ADD;	\
    mods[n]->mod_type = t; 		\
    mods[n++]->mod_values = v

    LDAPMod *mods[20];
    char new_dn[256], host[MAXHOSTNAMELEN];
    char *cn_v[] = {NULL, NULL};
    char *objectClass_v[] = {
	"top", "person", "organizationalPerson", "user", NULL 
    };
    char *samAccountName_v[] = {NULL, NULL};
    char *fullname = NULL, *office = "", *phone = "", *home_phone = "";
    char *p, *bp;
    char *name_v[] = {NULL, NULL};
    char *desc_v[] = {NULL, NULL};
    char *officeName_v[] = {NULL, NULL};
    char *telephoneNumber_v[] = {NULL, NULL};
    char *homePhone_v[] = {NULL, NULL};
    char upn[256];
    char *userPrincipalName_v[] = {NULL, NULL};
    char *userAccountControl_v[] = {NULL, NULL};
    char userAccountControlStr[80];
    char *posixHomeDirectory_v[] = {NULL, NULL};
    char *posixLoginShell_v[] = {NULL, NULL};
    char *posixUID_v[] = {NULL, NULL};
    char *posixGID_v[] = {NULL, NULL};
#ifdef HAVE_GETDOMAINNAME
    char *NISDomainName_v[] = {NULL, NULL};
    char nisdomain[256] = "";
#endif
    char uidStr[16], gidStr[16];
    char *description;
    u_int userAccountControl = UF_NORMAL_ACCOUNT;
    struct passwd *pwd;
    int n, rc, orc, i;

    if(invalidate)
      userAccountControl |= UF_ACCOUNTDISABLE;
    
    sprintf(upn, "%s@%s", user, domainname);

    samAccountName_v[0] = user;
    
    sprintf(userAccountControlStr, "%ld", userAccountControl);
    userAccountControl_v[0] = userAccountControlStr;
    userPrincipalName_v[0] = upn;

    /* XXX should free strdup()'ed items */
    pwd = getpwnam(user);
    if (pwd) {
	bp = pwd->pw_gecos;
	p = strtok(bp, ",");
	if (p) fullname = strdup(p);
	p = strtok(NULL, ",");
	if (p) office = strdup(p);
	p = strtok(NULL, ",");
	if (p) phone = strdup(p);
	p = strtok(NULL, ",");
	if (p) home_phone = strdup(p);
    }
    if (!fullname)
	fullname = strdup(user);
    
    cn_v[0] = fullname;
    name_v[0] = fullname;
    officeName_v[0] = office;
    telephoneNumber_v[0] = phone;
    homePhone_v[0] = home_phone;
    
    (void) gethostname(host, sizeof(host));
    description = malloc(strlen(host) + 80);
    if (!description)
	return ENOMEM;
    sprintf(description, "Account created from %s", host);
    desc_v[0] = description;
    
    sprintf(new_dn, "cn=%s,%s", fullname, bind_path);
    
    n = 0;
    ADD_ATTR("cn", cn_v);
    ADD_ATTR("objectClass", objectClass_v);
    ADD_ATTR("sAMAccountName", samAccountName_v);
    ADD_ATTR("userPrincipalName", userPrincipalName_v);
    ADD_ATTR("userAccountControl", userAccountControl_v);
    ADD_ATTR("name", name_v);
    ADD_ATTR("displayName", name_v);
    ADD_ATTR("description", desc_v);
    if (strlen(office)) {
	ADD_ATTR("physicalDeliveryOfficeName", officeName_v);
    }
    if (strlen(home_phone)) {
	ADD_ATTR("homePhone", homePhone_v);
    }
    if (strlen(phone)) {
	ADD_ATTR("telephoneNumber", telephoneNumber_v);
    }
    
    /* will replace any earlier settings */
    for (i=0; optattrNames[i]; i++) {
	char **values = malloc(2 * sizeof(char *));
	values[0] = optattrValues[i];
	values[1] = NULL;

	mods[n] = malloc(sizeof(LDAPMod));
	mods[n]->mod_op = LDAP_MOD_REPLACE;
	mods[n]->mod_type = optattrNames[i];
	mods[n++]->mod_values = values;
    }
    mods[n] = NULL;

    rc = ldap_add_ext_s(ld, new_dn, mods, NULL, NULL);
    free(description);
    if (rc != LDAP_SUCCESS)
	return rc;

    /* add optional Posix attributes */
    if (pwd) {
	posixHomeDirectory_v[0] = pwd->pw_dir;
	posixLoginShell_v[0] = pwd->pw_shell;
	sprintf(uidStr, "%d",  pwd->pw_uid);
	sprintf(gidStr, "%d",  pwd->pw_gid);
	posixUID_v[0] = uidStr;
	posixGID_v[0] = gidStr;
#ifdef HAVE_GETDOMAINNAME
	getdomainname(nisdomain, sizeof(nisdomain));
	NISDomainName_v[0] = nisdomain;
#endif    
    
	n = 0;
	ADD_ATTR("msSFUName", samAccountName_v);
#ifdef HAVE_GETDOMAINNAME
	if (strlen(nisdomain) > 0) {
	    ADD_ATTR("syncNisDomain", NISDomainName_v);
	}
#endif
	ADD_ATTR("msSFUHomeDirectory", posixHomeDirectory_v);
	ADD_ATTR("loginShell", posixLoginShell_v);
	ADD_ATTR("uidNumber", posixUID_v);
	ADD_ATTR("gidNumber", posixGID_v);
	mods[n] = NULL;

	orc = ldap_modify_ext_s(ld, new_dn, mods, NULL, NULL);
	if (orc != LDAP_SUCCESS) {
	    char *error_msg, *matched_msg;
	    fprintf(stderr, "WARNING: Failed to add Posix attributes: %s\n",
		    ldap_err2string(orc));
	    ldap_get_lderrno(ld, &matched_msg, &error_msg);
	    if (error_msg != NULL && *error_msg != '\0') {
		fprintf(stderr, "\t%s\n", error_msg);
	    }
	    if (matched_msg != NULL && *matched_msg != '\0') {
		fprintf(stderr, 
			"\tPart of the DN that matches an existing entry: %s\n", 
			matched_msg);
	    }
	    fprintf(stderr, "\tDirectory schema may not contain Posix schema.\n");
	}
    }
    else {
	fprintf(stderr, "WARNING: No local account in /etc/passwd for Posix information.\n");
    }
    
    return(rc);
}

int delete_user(
    LDAP *ld,
    char *del_dn
    )
{
    return ldap_delete_ext_s(ld, del_dn, NULL, NULL);
}

/*
**  Accepts null-terminated array of attributeName, attributeValue pairs.
**  If any values are not currently set, causes LDAP update.
**  Returns success of group update or success if no update.
*/

int add_string_values_to_attrs(
    LDAP *ld,
    LDAPMessage **res,
    char *bind_path,
    char *user,
    char **attributeNames,
    char **values
    )
{
    LDAPMod **mods;
    char ***modvalues;  /* each *mod has a char **modvalue */
    char user_dn[256];
    int i, j, changes, length;
    int retval=0;
    char **attrValues;

    /* Check for equal lengths */
    for (i=0; values[i]; i++);
    length = i;
    for (i=0; attributeNames[i]; i++);
    if (i != length)
	return LDAP_PARAM_ERROR;

    /* Check for unset attributes */
    /* Marshall a mod and modvalue for each unset */

    mods = malloc((length+1) * sizeof(LDAPMod *));
    modvalues = malloc((length+1) * sizeof(char **));

    for (i=0, changes=0; i < length; i++) {
	int numValues, foundValue = 0;
	attrValues = get_string_attr_value(ld, res, attributeNames[i], &numValues);

	if (attrValues) {
	    for (j=0; j < numValues && !foundValue; j++) {
		if (strcmp(attrValues[j], values[i]) == 0) {
		    ldap_value_free(attrValues);
		    foundValue = 1;
		    continue;                   /* already exists */
		}
	    }
	}

	if (!foundValue) {
	    ldap_value_free(attrValues);

	    /* doesn't exist, so marshall change */
	    mods[changes] = malloc(sizeof(LDAPMod));
	    modvalues[changes] = malloc(2 * sizeof(char *));

	    mods[changes]->mod_op = LDAP_MOD_ADD;
	    mods[changes]->mod_type = attributeNames[i];
	    modvalues[changes][0] = values[i];
	    modvalues[changes][1] = NULL;
	    mods[changes]->mod_values = modvalues[changes];
	    changes++;
	}
    }

    if (!changes) {
	free(mods);
	free(modvalues);
	return LDAP_SUCCESS;
    }
    else {
	mods[changes] = NULL;
    }

    sprintf(user_dn, "cn=%s,cn=users,%s", strdup(user), bind_path);
    retval = ldap_modify_ext_s( ld, user_dn, mods, NULL, NULL);

    /* free mods */
    for (i=0; i < changes; i++) {
	free(modvalues[i]);
	free(mods[i]);
    }
    free(mods);
    free(modvalues);

    return retval;
}

char **get_string_attr_value(
    LDAP *ld,
    LDAPMessage **res,
    char *attributeName,
    int *numVals
)
{
    char **attrValues;
    int i;

    attrValues = ldap_get_values(ld, *res, attributeName);
    *numVals = ldap_count_values(attrValues);

    return attrValues;
}

void list_attrs(
    LDAP *ld,
    LDAPMessage **res
    ) 
{
    BerElement *ber;
    char *attrs;
    int i;

    attrs = ldap_first_attribute(ld, *res, &ber);
    printf("attr[0] = %s\n", attrs);

    for (i=1; (attrs = ldap_next_attribute(ld, *res, ber)); i++) {
	printf("attrs[%i] = %s\n", i, attrs);
	ldap_memfree(attrs);
    }
}
