/*--

THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.

Copyright (C) 1999  Microsoft Corporation.  All rights reserved.

Module Name:

    adduser.c

Abstract:

    Sample program that demonstrates how to create a user account
    using LDAP.

--*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <krb5.h>
#include <lber.h>
#include <ldap.h>
#include "kpasswd.h"
#include "locator.h"
#include "person.h"

static const char rcsid[] = "$Id$";

#ifdef HAVE_LIBGSSLDAP
#include <gssldap.h>
#include <gsssasl.h>
#endif

char *argv0;
int verbose = 0;

struct creds 
{
    char *dn;
    char *pw;
} creds;

void dump_entry(LDAP *ld, LDAPMessage *res);
int convert_domain_to_dn(char *domain, char **bind_path);
int LDAP_CALL LDAP_CALLBACK ldap_simple_rebind( LDAP *ld, char **dnp, 
   char **passwdp, int *authmethodp, int freeit, void *arg );

void usage()
{
    printf("%s: [-vxI] [-A admin-account-dn] [-D domain] [-O ous] [-s LDAPserver] [-p LDAPport] [-B bind-path] [-a attributeName=value] <user@domain>\n", argv0);
    exit(1);
}

int main(int argc, char *argv[])
{
	LDAP *ld;
	int i, rc;
	LDAPMessage *res;
	char *bind_path = NULL;
	char *add_path = NULL;
	u_short ldap_port = LDAP_PORT;
	char *ldap_host = NULL;
	int ch;
	char *user, *cp;
	char *admindn = NULL;
	char *domain = NULL;
	char *defOU = "cn=users", *ou = NULL;
	int num_ds_hosts;
	PSRV_LOCATION *ds_hosts;
	int deluser = 0;
	int accountFound = 0;
	int invalidateAccount = 0;
	
	char *c;
	char *attributeNames[20];
	char *attributeValues[20];
	int numAttrs = 0;

	if (argv0 = strrchr(argv[0], '/'))
	    argv0++;
	else
	    argv0 = argv[0];
	
	while ((ch = getopt(argc, argv, "A:vxIs:p:a:B:D:O:?")) != -1) {
	    switch(ch) {
	    case 'v':
		verbose++;
		break;
	    case 'A':
		admindn = optarg;
		break;
	    case 'O':
		ou = optarg;
		break;
	    case 'D':
		domain = optarg;
		break;
	    case 'B':
		bind_path = optarg;
		break;
	    case 's':
		ldap_host = optarg;
		break;
	    case 'p':
		ldap_port = atoi(optarg);
		break;
	    case 'x':
		deluser = 1;
		break;
	    case 'I':
		invalidateAccount = 1;
		break;
	    case 'a':
	      if (!(c = strrchr(optarg, '=')))
		usage();
	      else {
		*c = '\0';
		attributeNames[numAttrs] = optarg;
		attributeValues[numAttrs++] = c + 1;
	      }		     
		break;
	    case '?':
	    default:
		usage();
	    }
	}

	argc -= optind;
	argv += optind;
	
	if (argc < 1) {
	    usage();
	}

	attributeNames[numAttrs] = NULL;
	attributeValues[numAttrs] = NULL;

	user = argv[0];

	if (!domain) {
	    domain = strchr(user, '@');
	    if (domain)
		domain++;
	    else
		usage();
	}

	if (verbose > 2)
	    ldap_debug = 1 /*LDAP_DEBUG_TRACE*/;

	if (cp = strchr(user, '@'))
	    *cp = '\0';

	if (!bind_path) {
	    convert_domain_to_dn(domain, &bind_path);
	    if (!ou)
	      ou = defOU;
	}
	
	if (verbose) {
	    if (!deluser)
	      printf("Creating user account for %s\n", user);
	    else
	      printf("Deleting user account for %s\n", user);
	}

	if (ldap_host) {
	    static SRV_LOCATION srv_ldap;
	    static PSRV_LOCATION psrv_ldap[1];

	    psrv_ldap[0] = &srv_ldap;
	    srv_ldap.srv_host = ldap_host;
	    srv_ldap.srv_port = ldap_port;
	    ds_hosts = psrv_ldap;
	    num_ds_hosts = 1;
	}
	else {
	    if (locate_srv(domain, LDAP_SRV, TCP_PROTO, NO_SITE, &ds_hosts, &num_ds_hosts) == -1 ||
		num_ds_hosts == 0) {
		printf("ERROR: Cannot locate a directory server for %s\n",
		       domain);
		exit(1);
	    }
	}
	
	for (i = 0; i < num_ds_hosts; i++) {
	    ld = ldap_open(ds_hosts[i]->srv_host, ds_hosts[i]->srv_port);
	    if (ld)
		break;
	}
	if (ld == NULL) {
	    perror("ldap_open");
	    return 1;
	}

	rc = LDAP_VERSION3;
	(void)ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &rc);

	if (admindn) {
	    char prompt[80];
	    char pw[1024];
	    int pwlen;
	    printf("Warning: Using simple LDAP bind.\n");
	    sprintf(prompt, "Password for %s:", admindn);
	    pwlen = sizeof(pw);
	    if (rc = krb5_read_password(NULL, prompt, NULL, pw, &pwlen)) {
		com_err(argv0, rc, "while reading admin password");
		return 1;
	    }
	    rc = ldap_simple_bind_s(ld, admindn, pw);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_simple_bind");
		ldap_unbind(ld);
		return 1;
	    }
	    creds.dn = strdup(admindn);
	    creds.pw = strdup(pw);
	    ldap_set_rebind_proc(ld, ldap_simple_rebind, &creds);
	    memset(pw, 0, sizeof(pw));
	}
	else {
#ifdef HAVE_LIBGSSLDAP
	    rc = ldap_gssapi_bind(ld, bind_path, GSSSASL_NO_SECURITY_LAYER);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_gssapi_bind");
		ldap_unbind(ld);
		return 1;
	    }
#else
	    printf("Warning: Using anonymous simple LDAP bind.\n");
	    rc = ldap_simple_bind_s(ld, NULL, NULL);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_simple_bind");
		ldap_unbind(ld);
		return 1;
	    }
#endif
	}

	if (ou) {
	    add_path = malloc(strlen(bind_path)+strlen(ou)+2); /*XXX*/
	    if (!add_path)
	        return ENOMEM;
	    sprintf(add_path, "%s,%s", ou, bind_path);
	}
	else
	  add_path = bind_path;   /*XXX malloc leak */
	
	if (verbose)
	    printf("Searching for \"%s\" at \"%s\" ...\n", user, bind_path);

	rc = find_user(ld, &res, bind_path, user);
	if (rc == LDAP_MORE_RESULTS_TO_RETURN ||
	    (rc == LDAP_SUCCESS && ldap_count_entries(ld, res) == 0)) {
	  if (!deluser) {
	    if (verbose)
		printf("Existing user account not found - adding as \"cn=%s,%s\"\n", user, add_path);
	    rc = add_user(ld, add_path, user, domain, invalidateAccount, 
			  attributeNames, attributeValues);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_add_s");
		ldap_unbind(ld);
		return 1;
	    }
	    rc = find_user(ld, &res, add_path, user);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_search_s");
		ldap_unbind(ld);
		return 1;
	    }
	  }
	  else {
	    if (verbose)
	      printf("Existing user account not found.\n");

	    return 1;
	  }
	}
	else {
	  accountFound = 1;

	  if (!deluser) {
	    if (verbose)
	      printf("User account already present\n");
	    
	    if (attributeNames[0])
	      rc = add_string_values_to_attrs(ld, &res, bind_path, user, 
					      attributeNames, attributeValues);
	    else
	      rc = LDAP_SUCCESS;
	    
	    if (rc != LDAP_SUCCESS) {
                ldap_perror(ld, "ldap_add_s");
                ldap_unbind(ld);
                return 1;
            }	    
	    if (attributeNames[0] && verbose) 
	      printf("Updates successful.\n");

	  }
	  else {
	    char *dn = ldap_get_dn(ld, res);
	    if (dn) {
		if (verbose)
		    printf("Existing user account found - deleting.\n");

		rc = delete_user(ld, dn);
		if (rc != LDAP_SUCCESS) {
		    ldap_memfree(dn);
		    ldap_perror(ld, "ldap_add_s");
		    ldap_unbind(ld);
		    return 1;
		}
		ldap_memfree(dn);
	    }
	    else
		printf("ERROR: User DN not known.\n");
	  }
	}

	if (verbose > 1)
	    dump_entry(ld, res);
	else
	    ldap_msgfree(res);
	ldap_unbind(ld);

	if (!invalidateAccount && !deluser && !accountFound)
	  rc = set_password(user, domain);

	return (rc != 0);
}

#define P1 "Enter new password: "
#define P2 "Enter it again: "

int set_password(
    char *user,
    char *domain
    )
{
    krb5_context context;
    krb5_data salt;
    krb5_encrypt_block eblock;
    krb5_keyblock * key;
    char pw[1024];
    int pwlen;
    int res_code;
    krb5_data res_code_str, res_string;
    krb5_principal user_princ, princ;
    krb5_error_code retval;
    krb5_ccache ccache;
    int i;
    char *userrealm, *realm;
    
    retval = krb5_init_context(&context);
    if (retval) {
	return retval;
    }
    
    if (retval = krb5_cc_default(context, &ccache)) {
 	com_err(argv0, retval, "opening default ccache");
	exit(1);
    }

    realm = malloc(strlen(domain)+1);		/*XXX*/
    if (!realm)
	return ENOMEM;
    
    for (i = 0; i < strlen(domain); i++)
	realm[i] = toupper(domain[i]);
    realm[i] = '\0';
    
    userrealm = malloc(strlen(user) + strlen(domain) + 2);	/*XXX*/
    if (!userrealm)
	return ENOMEM;
    sprintf(userrealm, "%s@%s", user, realm);
    retval = krb5_parse_name(context, userrealm, &user_princ);
    if (retval)
	return retval;
    
#define NUMTRIES  10

    if (verbose)
	printf("Setting password...\n");

    for (i = 0; i < NUMTRIES; i++) {
	pwlen = sizeof(pw);
	if (retval = krb5_read_password(context, P1, P2, pw, &pwlen)) {
	    com_err(argv0, retval, "while reading password");
	    continue;
	}
	retval = krb5_set_password(context, ccache, pw, user_princ,
				   &res_code, &res_code_str, &res_string);

	if (retval) {
	    com_err(argv0, retval, "setting user password");
	    exit(1);
	}
	if (res_code == KRB5_KPASSWD_SUCCESS)
	    break;
	else if (res_code != KRB5_KPASSWD_MALFORMED) {
	    printf("Set password for account failed: %.*s%s%.*s\n",
		   res_code_str.length, res_code_str.data,
		   res_string.length?": ":"",
		   res_string.length, res_string.data);
	    exit(2);
	}
	memset(pw, 0, sizeof(pw));
    }
    if (i == NUMTRIES)	/* Didn't get a good password */
	return KRB5KRB_AP_ERR_NOKEY;

    return retval;
}

int LDAP_CALL LDAP_CALLBACK ldap_simple_rebind(
    LDAP *ld,
    char **dnp,
    char **passwdp,
    int *authmethodp,
    int freeit,
    void *arg
    )
{
    struct creds *cp = (struct creds *)arg;
    
    if (!freeit) {
	*authmethodp = LDAP_AUTH_SIMPLE;
	*dnp = cp->dn;
	*passwdp = cp->pw;
    }
    return( LDAP_SUCCESS );
}

int convert_domain_to_dn(
    char *domain,
    char **dnp
    )
{
    char *fp, *dp;
    char dn[1024];
    int dnlen = 1;
    
    memset(dn, 0, sizeof(dn));
    strcpy(dn, "dc=");
    dp = dn+3;
    for (fp = domain; *fp; fp++) {
	if (*fp == '.') {
	    strcpy(dp, ",dc=");
	    dp += 4;
	}
	else
	    *dp++ = *fp;
    }

    *dnp = strdup(dn);
    return 0;
}

void dump_entry(
    LDAP *ld,
    LDAPMessage *res
    )
{
    LDAPMessage *e;
    
    for (e = ldap_first_entry(ld, res);
	 e;
	 e = ldap_next_entry(ld, e)) {

	BerElement *b;
	char *attr;
	char *dn = ldap_get_dn(ld, res);
	if (dn)
	    printf("dn: %s\n", dn);
	ldap_memfree(dn);

	for (attr = ldap_first_attribute(ld, e, &b);
	     attr;
	     attr = ldap_next_attribute(ld, e, b)) {

	    char **values, **p;
	    values = ldap_get_values(ld, e, attr);
	    for (p = values; *p; p++) {
		printf("%s: %s\n", attr, *p);
	    }
	    ldap_value_free(values);
	    ldap_memfree(attr);
	}

	ber_free(b, 1);
	printf("\n");
    }
}

