/*--

THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.

Copyright (C) 1999  Microsoft Corporation.  All rights reserved.

Module Name:

    kpasswd.c

Abstract:

    Change user password using the Kerberos Change Password Protocol (I-D)

--*/

#include <stdio.h>
#include <sys/types.h>
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <krb5.h>

static const char rcsid[] = "$Id$";

#define P1 "Enter new password: "
#define P2 "Enter it again: "

int main(int argc, char *argv[])
{
   krb5_error_code ret;
   krb5_context context;
   krb5_principal princ;
   char *pname;
#ifdef HAVE_PWD_H
   struct passwd *pwd;
#endif
   krb5_ccache ccache;
   krb5_timestamp now;
   char *client_name, prompt[255];
   krb5_principal server;
   krb5_creds creds;
   char pw[1024];
   int pwlen;
   int result_code;
   krb5_data result_code_string, result_string;
   
   memset(&result_string, 0, sizeof(result_string));
   
   if (argc > 2) {
      fprintf(stderr, "usage: %s [principal]\n", argv[0]);
      exit(1);
   }

   pname = argv[1];

   if (ret = krb5_init_context(&context)) {
      com_err(argv[0], ret, "initializing kerberos library");
      exit(1);
   }

   if ((ret = krb5_timeofday(context, &now))) {
      com_err(argv[0], ret, "while getting time of day");
      exit(1);
   }

   /* in order, use the first of:
      - a name specified on the command line
      - the principal name from an existing ccache
      - the name corresponding to the ruid of the process

      otherwise, it's an error.
      */

   if (pname) {
      if (ret = krb5_parse_name(context, pname, &princ)) {
	 com_err(argv[0], ret, "parsing client name");
	 exit(1);
      }
   } else if ((ret = krb5_cc_default(context, &ccache)) != KRB5_CC_NOTFOUND) {
      if (ret) {
	 com_err(argv[0], ret, "opening default ccache");
	 exit(1);
      }

      if (ret = krb5_cc_get_principal(context, ccache, &princ)) {
	 com_err(argv[0], ret, "getting principal from ccache");
	 exit(1);
      }

      if (ret = krb5_cc_close(context, ccache)) {
	 com_err(argv[0], ret, "closing ccache");
	 exit(1);
      }
#ifdef HAVE_PWD_H
   } else if (pwd = getpwuid(getuid())) {
      if (ret = krb5_parse_name(context, pwd->pw_name, &princ)) {
	 com_err(argv[0], ret, "parsing client name");
	 exit(1);
      }
#endif
   } else {
      com_err(argv[0], 0,
	      "no matching password entry while looking for username");
      exit(1);
   }

    memset((char *)&creds, 0, sizeof(creds));
    creds.client = princ;

    if (ret = krb5_parse_name(context, "kadmin/changepw", &server)) {
	com_err(argv[0], ret, "while parsing service name");
	exit(1);
    }
    krb5_princ_set_realm(context, server, krb5_princ_realm(context, princ));

    creds.server = server;

    creds.times.starttime = 0;
    creds.times.endtime = now + (5*60);

    if (ret = krb5_unparse_name(context, princ, &client_name)) {
	com_err(argv[0], ret, "while parsing client name");
	exit(1);
    }

    (void) sprintf(prompt,"Password for %s: ", (char *) client_name);

    pwlen = sizeof(pw);

    ret = krb5_read_password(context, prompt, 0, pw, &pwlen);
    if (ret || pwlen == 0) {
       fprintf(stderr, "Error while reading password for '%s'\n",
	       client_name);
       memset(pw, 0, sizeof(pw));
       exit(1);
    }

    if (ret = krb5_get_in_tkt_with_password(context, 0, NULL,
					    NULL, NULL, pw, 0,
					    &creds, 0)) {
       memset(pw, 0, sizeof(pw));
       if (ret == KRB5KRB_AP_ERR_BAD_INTEGRITY)
	   com_err(argv[0], 0,
		   "Password incorrect while getting initial ticket");
       else
	   com_err(argv[0], ret, "getting initial ticket");
       exit(1);
    }
    memset(pw, 0, sizeof(pw));

   pwlen = sizeof(pw);
   if (ret = krb5_read_password(context, P1, P2, pw, &pwlen)) {
      com_err(argv[0], ret, "while reading password");
      exit(1);
   }

   if (ret = krb5_change_password(context, &creds, pw,
				  &result_code, &result_code_string,
				  &result_string)) {
      com_err(argv[0], ret, "changing password");
      exit(1);
   }

   if (result_code) {
      printf("%.*s%s%.*s\n",
	     result_code_string.length, result_code_string.data,
	     result_string.length?": ":"",
	     result_string.length,
	     result_string.length?result_string.data:"");
      exit(2);
   }

   free(result_string.data);
   free(result_code_string.data);

   printf("Password changed.\n");
   exit(0);
}
