/*--

THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.

Copyright (C) 1999  Microsoft Corporation.  All rights reserved.

Module Name:

    netjoin.c

Abstract:

    Sample program that demonstrates how to create a computer account
    using LDAP.

--*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/param.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <krb5.h>
#include <lber.h>
#include <ldap.h>
#include "kpasswd.h"
#include "locator.h"
#include "computer.h"

static const char rcsid[] = "$Id$";

#ifdef HAVE_LIBGSSLDAP
#include <gssldap.h>
#include <gsssasl.h>
#endif

char *argv0;
int verbose = 0;

struct creds 
{
    char *dn;
    char *pw;
} creds;

void dump_entry(LDAP *ld, LDAPMessage *res);
int convert_domain_to_dn(char *domain, char **bind_path);
int LDAP_CALL LDAP_CALLBACK ldap_simple_rebind( LDAP *ld, char **dnp, 
   char **passwdp, int *authmethodp, int freeit, void *arg );

void usage()
{
    printf("%s: [-v] [-s LDAPserver] [-D domain] [-O ous] [-p LDAPport] [-A admin-account-dn] [-B bind-path] [host name]\n", argv0);
    exit(1);
}

int main(int argc, char *argv[])
{
	LDAP *ld;
	int i, rc;
	LDAPMessage *res;
	char *bind_path = NULL;
	char *add_path = NULL;
	u_short ldap_port = LDAP_PORT;
	char *ldap_host = NULL;
	char *cp;
	int ch;
	char *host, *hostname = NULL;
	char *admindn = NULL;
	char *domain = NULL;
	char *defOU = "cn=computers", *ou = NULL;
	int num_ds_hosts;
	PSRV_LOCATION *ds_hosts;
	struct hostent *hp;

	if (argv0 = strrchr(argv[0], '/'))
	    argv0++;
	else
	    argv0 = argv[0];
	
	while ((ch = getopt(argc, argv, "vs:p:A:B:D:O:?")) != -1) {
	    switch(ch) {
	    case 'v':
		verbose++;
		break;
	    case 'A':
		admindn = optarg;
		break;
	    case 'O':
		ou = optarg;
		break;
	    case 'D':
		domain = optarg;
		break;
	    case 'B':
		bind_path = optarg;
		break;
	    case 's':
		ldap_host = optarg;
		break;
	    case 'p':
		ldap_port = atoi(optarg);
		break;
	    case '?':
	    default:
		usage();
	    }
	}

	argc -= optind;
	argv += optind;
	
	if (argc < 1) {
	    host = malloc(MAXHOSTNAMELEN);
	    if (!host) {
		fprintf(stderr, "Out of memory\n");
		exit(1);
	    }
	    gethostname(host, MAXHOSTNAMELEN);
	}
	else
	    host = argv[0];

	if (verbose > 2)
	    ldap_debug = 1 /*LDAP_DEBUG_TRACE*/;

	hp = gethostbyname(host);
	if (hp) {
	    hostname = strdup(hp->h_name);
	}

	if (!hostname) {
	    fprintf(stderr, "Unable to resolve hostname - %s\n", host);
	    exit(1);
	}

	if (cp = strchr(host, '.'))
	    *cp = '\0';

	if (!domain) {
	    domain = strdup(hp->h_name);
	    domain = strchr(domain, '.');
	    if (domain)
		domain++;
            if (!domain) {
                fprintf(stderr, "ERROR: No domain for host \"%s\"\n", hp->h_nam\
e);
                exit(1);
            }
	}

	if (!bind_path) {
	    convert_domain_to_dn(domain, &bind_path);
	    if (!ou)
	      ou = defOU;
	}
	
	if (verbose)
	    printf("Creating host account for %s@%s\n", hostname, domain);

	if (ldap_host) {
	    static SRV_LOCATION srv_ldap;
	    static PSRV_LOCATION psrv_ldap[1];

	    psrv_ldap[0] = &srv_ldap;
	    srv_ldap.srv_host = ldap_host;
	    srv_ldap.srv_port = ldap_port;
	    ds_hosts = psrv_ldap;
	    num_ds_hosts = 1;
	}
	else {
	    if (locate_srv(domain, LDAP_SRV, TCP_PROTO, NO_SITE, &ds_hosts, &num_ds_hosts) == -1 ||
		num_ds_hosts == 0) {
		printf("ERROR: Cannot locate a directory server for %s\n",
		       domain);
		exit(1);
	    }
	}
	
	for (i = 0; i < num_ds_hosts; i++) {
	    ld = ldap_open(ds_hosts[i]->srv_host, ds_hosts[i]->srv_port);
	    if (ld)
		break;
	}
	if (ld == NULL) {
	    perror("ldap_open");
	    return 1;
	}

	rc = LDAP_VERSION3;
	(void)ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &rc);

	if (admindn) {
	    char prompt[80];
	    char pw[1024];
	    int pwlen;
	    printf("Warning: Using simple LDAP bind.\n");
	    sprintf(prompt, "Password for %s:", admindn);
	    pwlen = sizeof(pw);
	    if (rc = krb5_read_password(NULL, prompt, NULL, pw, &pwlen)) {
		com_err(argv0, rc, "while reading admin password");
		return 1;
	    }
	    rc = ldap_simple_bind_s(ld, admindn, pw);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_simple_bind");
		ldap_unbind(ld);
		return 1;
	    }
	    creds.dn = strdup(admindn);
	    creds.pw = strdup(pw);
	    ldap_set_rebind_proc(ld, ldap_simple_rebind, &creds);
	    memset(pw, 0, sizeof(pw));
	}
	else {
#ifdef HAVE_LIBGSSLDAP
	    rc = ldap_gssapi_bind(ld, bind_path, GSSSASL_NO_SECURITY_LAYER);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_gssapi_bind");
		ldap_unbind(ld);
		return 1;
	    }
#else
	    printf("Warning: Using anonymous simple LDAP bind.\n");
	    rc = ldap_simple_bind_s(ld, NULL, NULL);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_simple_bind");
		ldap_unbind(ld);
		return 1;
	    }
#endif
	}

	if (ou) {
	    add_path = malloc(strlen(bind_path)+strlen(ou)+2); /*XXX*/
	    if (!add_path)
	        return ENOMEM;
	    sprintf(add_path, "%s,%s", ou, bind_path);
	}
	else
	    add_path = bind_path;   /*XXX malloc leak */
	
	if (verbose)
	    printf("Searching for \"%s\" at \"%s\" ...\n", host, bind_path);

	rc = find_host(ld, &res, bind_path, host);
	if (rc == LDAP_MORE_RESULTS_TO_RETURN ||
	    (rc == LDAP_SUCCESS && ldap_count_entries(ld, res) == 0)) {
	    if (verbose)
		printf("Existing host account not found - adding as \"cn=%s,%s\"\n", host, add_path);
	    rc = add_host(ld, add_path, host, hostname, domain, (argc == 0));
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_add_s");
		ldap_unbind(ld);
		return 1;
	    }
	    rc = find_host(ld, &res, bind_path, host);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_search_s");
		ldap_unbind(ld);
		return 1;
	    }
	}
	else
	    if (verbose)
		printf("Host account already present\n");

	if (verbose > 1)
	    dump_entry(ld, res);
	else
	    ldap_msgfree(res);
	ldap_unbind(ld);

	rc = set_password(hostname, domain, (argc == 0));
	return (rc != 0);
}
	
/*
 * Generate a simple random password of 15 chars - not a cryptographic one
 */
int generate_random_password(
    krb5_data *pwd
    )
{
    int i;
    int punct = 0;
#define PWLEN 15

    if (!(pwd->data = malloc(PWLEN+1)))
	return ENOMEM;
    pwd->length = PWLEN;
    pwd->magic = KV5M_PWD_DATA;

    memset(pwd->data, 0, pwd->length+1);
    
    srandom((unsigned)time(NULL));
    while (!punct) {
	i = rand() % 0x7f;
	if (ispunct(i))
	    punct = i;
    }

    for (i = 0; i <= PWLEN; i++) {
	if (punct && rand() & 0x100) {
	    pwd->data[i] = punct;
	    punct = 0;
	}
	else
	    while(!isalnum(pwd->data[i]) || !isprint(pwd->data[i])) {
		pwd->data[i] = random() & 0x7f;
	    }
    }
    pwd->data[i] = '\0';
    pwd->length = strlen(pwd->data);
    return 0;
}

int set_password(
    char *host,
    char *domain,
    int use_default_kt
    )
{
    krb5_context context;
    krb5_data salt;
    krb5_encrypt_block eblock;
    krb5_keyblock * key;
    krb5_data password;
    int res_code;
    krb5_data res_code_str, res_string;
    krb5_keytab kt;
    krb5_keytab_entry entry, *ent;
    krb5_kt_cursor cursor;
    krb5_principal host_princ, princ, server;
    krb5_error_code retval;
    krb5_ccache ccache;
    int kvno = 1, i;
    char *hostrealm, *realm;
    char ktname[MAXPATHLEN+sizeof("WRFILE:")+1];
    extern krb5_kt_ops krb5_ktf_writable_ops;
    
    retval = krb5_init_context(&context);
    if (retval) {
	return retval;
    }
    
    if (retval = krb5_cc_default(context, &ccache)) {
 	com_err(argv0, retval, "opening default ccache");
	exit(1);
    }

    realm = malloc(strlen(domain)+1);
    if (!realm)
	return ENOMEM;
    
    for (i = 0; i < strlen(domain); i++)
	realm[i] = toupper(domain[i]);
    realm[i] = '\0';
    
    hostrealm = malloc(strlen(host) + strlen(domain) + 8);
    sprintf(hostrealm, "host/%s@%s", host, realm);
    free(realm);
    retval = krb5_parse_name(context, hostrealm, &host_princ);
    free(hostrealm);
    if (retval)
	return retval;
    
#define NUMTRIES  10

    if (verbose)
	printf("Setting computer password...\n");

    for (i = 0; i < NUMTRIES; i++) {
	generate_random_password(&password);
	retval = krb5_set_password(context, ccache, password.data, host_princ,
				   &res_code, &res_code_str, &res_string);

	if (retval) {
	    com_err(argv0, retval, "setting computer password");
	    exit(1);
	}
	if (res_code == KRB5_KPASSWD_SUCCESS)
	    break;
	else if (res_code != KRB5_KPASSWD_MALFORMED) {
	    printf("Set password for computer account failed: %.*s%s%.*s\n",
		   res_code_str.length, res_code_str.data,
		   res_string.length?": ":"",
		   res_string.length, res_string.data);
	    exit(2);
	}
	memset(password.data, 0, password.length);
	free(password.data);
    }
    if (i == NUMTRIES)	/* Can't generate a good random password */
	return KRB5KRB_AP_ERR_NOKEY;

    retval = krb5_principal2salt(context, host_princ, &salt);
    if (retval)
	return retval;
    
    if (!(key = (krb5_keyblock *)malloc(sizeof(*key))))
	return ENOMEM;

    krb5_use_enctype(context, &eblock, ENCTYPE_DES_CBC_MD5);
    
    retval = krb5_string_to_key(context, &eblock, key, &password, &salt);
    if (retval)
	return retval;

    if (verbose > 2) {
	printf("Host password \"%s\" key ", password.data);
	printf(" (0x");
	{
	    int i;
	    for (i = 0; i < key->length; i++)
		printf("%02x", key->contents[i]);
	}
	printf(")\n");
    }

    memset(password.data, 0, password.length);

    retval = krb5_kt_register(context, &krb5_ktf_writable_ops);
    if (retval) {
	com_err(argv0, retval,
		"while registering writable key table functions");
	exit(1);
    }

    if (use_default_kt) {
	krb5_kt_default_name(context, &ktname[2], sizeof(ktname)-2);
	ktname[0] = 'W';
	ktname[1] = 'R';
    }
    else {
	char *shost = strdup(host), *p;
	p = strchr(shost, '.');
	if (p) *p = '\0';
	sprintf(ktname, "WRFILE:/tmp/%s.keytab", shost);
	free(shost);
    }

    if (verbose)
	printf("default keytab name - \"%s\"\n", ktname);

    retval = krb5_kt_resolve(context, ktname, &kt);
    if (retval) {
 	com_err(argv0, retval, "resolving keytab");
	return retval;
    }

    retval = krb5_kt_start_seq_get(context, kt, &cursor);
    if (retval == 0) {
	for (;;) {
	    retval = krb5_kt_next_entry(context, kt, &entry, &cursor);
	    if (retval)
		break;
	    if (krb5_principal_compare(context, entry.principal, host_princ)) {
		if (entry.vno > kvno)
		    kvno = entry.vno;
	    }
	}
 	if (retval == KRB5_KT_END)
	    retval = 0;
	krb5_kt_end_seq_get(context, kt, &cursor);
	kvno++;
    }

    entry.magic = KV5M_KEYTAB;
    entry.principal = host_princ;
    entry.vno = kvno;
    entry.key = *key;
    if (verbose)
	printf("Saving key version %d in %s\n", kvno, ktname);
    retval = krb5_kt_add_entry(context, kt, &entry);
    if (retval)
 	com_err(argv0, retval, "writing keytab");

    krb5_kt_close(context, kt);

    return retval;
}

int LDAP_CALL LDAP_CALLBACK ldap_simple_rebind(
    LDAP *ld,
    char **dnp,
    char **passwdp,
    int *authmethodp,
    int freeit,
    void *arg
    )
{
    struct creds *cp = (struct creds *)arg;
    
    if (!freeit) {
	*authmethodp = LDAP_AUTH_SIMPLE;
	*dnp = cp->dn;
	*passwdp = cp->pw;
    }
    return( LDAP_SUCCESS );
}

int convert_domain_to_dn(
    char *domain,
    char **dnp
    )
{
    /*
     * Change: mit.nttest.microsoft.com to
     *         dc=mit,dc=nttest,dc=microsoft,dc=com
     */
    char *fp, *dp;
    char dn[1024];
    int dnlen = 1;
    
    memset(dn, 0, sizeof(dn));
    strcpy(dn, "dc=");
    dp = dn+3;
    for (fp = domain; *fp; fp++) {
	if (*fp == '.') {
	    strcpy(dp, ",dc=");
	    dp += 4;
	}
	else
	    *dp++ = *fp;
    }

    *dnp = strdup(dn);
    return 0;
}


void dump_entry(
    LDAP *ld,
    LDAPMessage *res
    )
{
    LDAPMessage *e;
    
    for (e = ldap_first_entry(ld, res);
	 e;
	 e = ldap_next_entry(ld, e)) {

	BerElement *b;
	char *attr;
	char *dn = ldap_get_dn(ld, res);
	if (dn)
	    printf("dn: %s\n", dn);
	ldap_memfree(dn);

	for (attr = ldap_first_attribute(ld, e, &b);
	     attr;
	     attr = ldap_next_attribute(ld, e, b)) {

	    char **values, **p;
	    values = ldap_get_values(ld, e, attr);
	    for (p = values; *p; p++) {
		printf("%s: %s\n", attr, *p);
	    }
	    ldap_value_free(values);
	    ldap_memfree(attr);
	}

	ber_free(b, 1);
	printf("\n");
    }
}

