/*--

THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.

Copyright (C) 1999  Microsoft Corporation.  All rights reserved.

Module Name:

    pwdump.c

Abstract:

    Sample program that demonstrates how to dump a user account into a pw
    struct using LDAP.

--*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <krb5.h>
#include <lber.h>
#include <ldap.h>
#include "kpasswd.h"
#include "locator.h"
#include "person.h"
#include "sam.h"

static const char rcsid[] = "$Id$";

#ifndef DEF_SHELL
#define DEF_SHELL "/bin/sh"
#endif

#ifdef HAVE_LIBGSSLDAP
#include <gssldap.h>
#include <gsssasl.h>
#endif

char *argv0;
int verbose = 0;

struct creds 
{
    char *dn;
    char *pw;
} creds;

void dump_entry(LDAP *ld, LDAPMessage *res);
int dump_pw(LDAP *ld, LDAPMessage *res);
int convert_domain_to_dn(char *domain, char **bind_path);
int LDAP_CALL LDAP_CALLBACK ldap_simple_rebind( LDAP *ld, char **dnp, 
   char **passwdp, int *authmethodp, int freeit, void *arg );

void usage()
{
    printf("%s: [-v] [-A admin-account-dn] [-D domain] [-s LDAPserver] [-p LDAPport] [-B bind-path] [-O ou] <user@domain>\n", argv0);
    exit(1);
}

int main(int argc, char *argv[])
{
	LDAP *ld;
	int i, rc;
	LDAPMessage *res;
	char *bind_path = NULL;
	char *search_path;
	u_short ldap_port = LDAP_PORT;
	char *ldap_host = NULL;
	int ch;
	char *user, *cp, *ou = NULL;
	char *admindn = NULL;
	char *domain = NULL;
	int num_ds_hosts, status = 0;
	PSRV_LOCATION *ds_hosts;

	if (argv0 = strrchr(argv[0], '/'))
	    argv0++;
	else
	    argv0 = argv[0];
	
	while ((ch = getopt(argc, argv, "A:vIs:p:a:B:D:O:?")) != -1) {
	    switch(ch) {
	    case 'v':
		verbose++;
		break;
	    case 'A':
		admindn = optarg;
		break;
	    case 'D':
		domain = optarg;
		break;
	    case 'B':
		bind_path = optarg;
		break;
	    case 'O':
		ou = optarg;
		break;
	    case 's':
		ldap_host = optarg;
		break;
	    case 'p':
		ldap_port = atoi(optarg);
		break;

	    case '?':
	    default:
		usage();
	    }
	}

	argc -= optind;
	argv += optind;
	
	if (argc < 1) {
	    usage();
	}

	user = argv[0];

	if (!domain) {
	    domain = strchr(user, '@');
	    if (domain)
		domain++;
	    else
		usage();
	}

	if (verbose > 2)
	    ldap_debug = 1 /*LDAP_DEBUG_TRACE*/;
	
	if (cp = strchr(user, '@'))
	    *cp = '\0';

	if (!bind_path)
	    convert_domain_to_dn(domain, &bind_path);

	if (ldap_host) {
	    static SRV_LOCATION srv_ldap;
	    static PSRV_LOCATION psrv_ldap[1];

	    psrv_ldap[0] = &srv_ldap;
	    srv_ldap.srv_host = ldap_host;
	    srv_ldap.srv_port = ldap_port;
	    ds_hosts = psrv_ldap;
	    num_ds_hosts = 1;
	}
	else {
	    if (locate_srv(domain, LDAP_SRV, TCP_PROTO, NO_SITE, &ds_hosts, &num_ds_hosts) == -1 ||
		num_ds_hosts == 0) {
		printf("ERROR: Cannot locate a directory server for %s\n",
		       domain);
		exit(1);
	    }
	}
	
	for (i = 0; i < num_ds_hosts; i++) {
	    if (verbose)
		printf("Connecting to %s ...\n", ds_hosts[i]->srv_host);
	    ld = ldap_open(ds_hosts[i]->srv_host, ds_hosts[i]->srv_port);
	    if (ld)
		break;
	}
	if (ld == NULL) {
	    perror("ldap_open");
	    exit(1);
	}
	if (verbose)
	    printf("connected\n");
	
	rc = LDAP_VERSION3;
	(void)ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &rc);

	if (admindn) {
	    char prompt[80];
	    char pw[1024];
	    int pwlen;
	    printf("Warning: Using simple LDAP bind.\n");
	    sprintf(prompt, "Password for %s:", admindn);
	    pwlen = sizeof(pw);
	    if (rc = krb5_read_password(NULL, prompt, NULL, pw, &pwlen)) {
		com_err(argv0, rc, "while reading admin password");
		exit(1);
	    }
	    rc = ldap_simple_bind_s(ld, admindn, pw);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_simple_bind");
		ldap_unbind(ld);
		exit(1);
	    }
	    creds.dn = strdup(admindn);
	    creds.pw = strdup(pw);
	    ldap_set_rebind_proc(ld, ldap_simple_rebind, &creds);
	    memset(pw, 0, sizeof(pw));
	}
	else {
#ifdef HAVE_LIBGSSLDAP
	    rc = ldap_gssapi_bind(ld, bind_path, GSSSASL_NO_SECURITY_LAYER);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_gssapi_bind");
		ldap_unbind(ld);
		exit(1);
	    }
#else
	    printf("Warning: Using anonymous simple LDAP bind.\n");
	    rc = ldap_simple_bind_s(ld, NULL, NULL);
	    if (rc != LDAP_SUCCESS) {
		ldap_perror(ld, "ldap_simple_bind");
		ldap_unbind(ld);
		exit(1);
	    }
#endif
	}
	
	if (ou) {
	    search_path = malloc(strlen(ou) + strlen(bind_path) + 2);
	    if (!search_path) {
		fprintf(stderr, "Malloc failed.\n");
		exit(2);
	    }
	    sprintf(search_path, "%s,%s", ou, bind_path);
	}
	else
	    search_path = bind_path;
	
	if (verbose)
	    printf("Searching for \"%s\" in \"%s\"\n", user, search_path);
	
	rc = find_user(ld, &res, search_path, user);
	if (rc == LDAP_SUCCESS) {
	    if (verbose > 1)
		dump_entry(ld, res);

	    if (!dump_pw(ld, res)) {
		status = 1;
	    }
	}
	else
	    status = 1;
	
	if (status != 0)
	    printf("No entry found for %s\n", user);
	
	ldap_unbind(ld);

	exit(status);
}

int LDAP_CALL LDAP_CALLBACK ldap_simple_rebind(
    LDAP *ld,
    char **dnp,
    char **passwdp,
    int *authmethodp,
    int freeit,
    void *arg
    )
{
    struct creds *cp = (struct creds *)arg;
    
    if (!freeit) {
	*authmethodp = LDAP_AUTH_SIMPLE;
	*dnp = cp->dn;
	*passwdp = cp->pw;
    }
    return(LDAP_SUCCESS);
}


int convert_domain_to_dn(
    char *domain,
    char **dnp
    )
{
    char *fp, *dp;
    char dn[1024];
    int dnlen = 1;
    
    memset(dn, 0, sizeof(dn));
    strcpy(dn, "dc=");
    dp = dn+3;
    for (fp = domain; *fp; fp++) {
	if (*fp == '.') {
	    strcpy(dp, ",dc=");
	    dp += 4;
	}
	else
	    *dp++ = *fp;
    }

    *dnp = strdup(dn);
    return 0;
}

void dump_entry(
    LDAP *ld,
    LDAPMessage *res
    )
{
    LDAPMessage *e;
    
    for (e = ldap_first_entry(ld, res);
	 e;
	 e = ldap_next_entry(ld, e)) {

	BerElement *b;
	char *attr;
	char *dn = ldap_get_dn(ld, res);
	if (dn)
	    printf("dn: %s\n", dn);
	ldap_memfree(dn);

	for (attr = ldap_first_attribute(ld, e, &b);
	     attr;
	     attr = ldap_next_attribute(ld, e, b)) {

	    char **values, **p;
	    values = ldap_get_values(ld, e, attr);
	    for (p = values; *p; p++) {
		printf("%s: %s\n", attr, *p);
	    }
	    ldap_value_free(values);
	    ldap_memfree(attr);
	}

	ber_free(b, 1);
	printf("\n");
    }
}

int dump_pw(
    LDAP *ld,
    LDAPMessage *res
    )
{
    LDAPMessage *e;
    char *fn = NULL, *office = NULL, *homeph = NULL, *phone = NULL;
    char *pw_name = NULL, *pw_shell = NULL, *pw_dir = NULL;
    u_int userAccountControl;
    int pw_uid, pw_gid;
    int found_one = FALSE;
    
    for (e = ldap_first_entry(ld, res);
	 e;
	 e = ldap_next_entry(ld, e)) {

	BerElement *b;
	char *attr;

#define DSTR(s, d) 		\
	    (s)?s:d

#define FREE_STR(s)		\
	    if (s) {		\
		free(s);	\
		s = NULL;	\
	    }

	pw_uid = -1;
	pw_gid = -1;
	userAccountControl = 0L;

	FREE_STR(pw_dir);
	FREE_STR(pw_shell);
	FREE_STR(fn);
	FREE_STR(office);
	FREE_STR(homeph);
	FREE_STR(phone);

    
	for (attr = ldap_first_attribute(ld, e, &b);
	     attr;
	     attr = ldap_next_attribute(ld, e, b)) {

	    char **values, **p;
	    values = ldap_get_values(ld, e, attr);
	    for (p = values; *p; p++) {
		if (!strcmp(attr, "sAMAccountName"))
		    pw_name = strdup(*p);
		else if (!strcmp(attr, "uidNumber"))
		    pw_uid = atoi(*p);
		else if (!strcmp(attr, "gidNumber"))
		    pw_gid = atoi(*p);
		else if (!strcmp(attr, "displayName"))
		    fn = strdup(*p);
		else if (!strcmp(attr, "physicalDeliveryOfficeName"))
		    office = strdup(*p);
		else if (!strcmp(attr, "telephoneNumber"))
		    phone = strdup(*p);
		else if (!strcmp(attr, "homePhone"))
		    homeph = strdup(*p);
		else if (!strcmp(attr, "msSFUHomeDirectory"))
		    pw_dir = strdup(*p);
		else if (!strcmp(attr, "loginShell"))
		    pw_shell = strdup(*p);
		else if (!strcmp(attr, "userAccountControl"))
		    userAccountControl = atoi(*p);
	    }
	    ldap_value_free(values);
	    ldap_memfree(attr);
	}

	if (userAccountControl & UF_NORMAL_ACCOUNT) {
	    printf("%s:%s:%d:%d:%s,%s,%s,%s:%s:%s\n",
		   DSTR(pw_name, ""), "*", pw_uid, pw_gid,
		   DSTR(fn, ""), DSTR(office, ""), DSTR(phone, ""),
		   DSTR(homeph, ""), DSTR(pw_dir, "/"),
		   DSTR(pw_shell, DEF_SHELL));
	    found_one = TRUE;
	}
	else if (userAccountControl && verbose) {
	    printf("%s: not a user account (%x)\n",
		   pw_name, userAccountControl);
	}

	ber_free(b, 1);
    }
    return (found_one);
}
